#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <signal.h>
#include <getopt.h>
#include <errno.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <ul_htimer.h>
#include <can_vca.h>
#include <vca_fdpoll.h>
#include <vca_custlog.h>
#include <vca_pollfdnodecan.h>
#include <vca_net.h>

#ifdef WITH_DAEMON
#include <syslog.h>
#endif /*WITH_DAEMON*/


#include "canblaster_defaults.h"
#include "socknode.h"

#define mkstr(a) mkstr1(a)
#define mkstr1(a) # a

VCA_DECLARE_CUSTOM_LOG(canblaster)

//=============================================================
static char *can_device_name = "/dev/can0";
static int server_port = CANBLASTER_DEFAULT_PORT;
static FdNodeSrvSock_t server_node;
static vcaNet_t can_net;

static char help_msg[] =
            "USAGE:\n"
            "%s [OPTION]\n\n"
			"OPTIONS:\n"
			"\t-h\n"
            "\t--help\tthis help screen\n"
            "\t-d\n"
            "\t--candev\tCAN device name (default /dev/can0)\n"
            "\t-p\n"
            "\t--port\tIP port number on which server will listen to. Default is "mkstr(CANBLASTER_DEFAULT_PORT)"\n"
			"\t-g\n"
            "\t--terminate-from-tty\taccept terminate request from standard input.\n"
    #ifdef WITH_DAEMON
            "\t--daemon\tdaemonize\n"
    #endif /*WITH_DAEMON*/
            "\t--log_level [n]\tsets how many log messages you will see.\n"
            "\t\t0 - fatal errors only\n"
            "\t\t1 - level 0 + errors\n"
            "\t\t2 - level 1 + messages\n"
            "\t\t3 - level 2 + info messages\n"
            "\t\t4 - level 3 + debug messages\n"
            "\t\tlog level can be also set by environment variable CANMOND_LOG_LEVEL\n"
			"\t-v --verbose same as --log_level 3\n";
//--------------------------------------------------------------------
void clean_up(void)
{
    // close server socket
    if(server_node.fdnode.fd > 0) {
        clog(LOG_DEB, "closing server socket\n");
        close(server_node.fdnode.fd);
    }
    // TODO: check code to close CAN communication
    
    FdNodeSrvSock_destroy(&server_node);

    vcaNet_destroy(&can_net);
    
    /*
    if(can_node.fdnode.fd > 0) {
        log(LOG_DEB, "closing CAN device\n");
        close(can_node.fdnode.fd);
    }
    */
    /*
    // remove all clients
    for(item=client_first(&client_root); item; item = client_next(&client_root, item)) {
        delete_client(item->fd);
    }
    */
}

//----------------------------------------------------
/* Support for termination from stdin*/

typedef struct cmdttyFdNode_t
{
   vcaFdNode_t fdnode;
   int *terminate_ptr;
} cmdttyFdNode_t;


void cmdttyFdReady(struct vcaFdNode_t *nd, int event_type, struct vcaFdPoll_t *p)
{
    cmdttyFdNode_t *ttynd = UL_CONTAINEROF(nd, cmdttyFdNode_t, fdnode);
    *(ttynd->terminate_ptr) = 1;
}


void cmdttyFdNode_init(cmdttyFdNode_t *ttynd, vca_fd_t fd, int *terminate_ptr)
{
    if(ttynd == NULL) {
        ttynd = (cmdttyFdNode_t*)malloc(sizeof(cmdttyFdNode_t));
        if(!ttynd) fatal_error("memory allocation ERROR\n");
    }
    vcaFdNode_init(&(ttynd->fdnode), fd, VCAFD_RD);
    ttynd->terminate_ptr = terminate_ptr;
    ttynd->fdnode.ready = cmdttyFdReady;
}

cmdttyFdNode_t cmdttyFdNode;

//----------------------------------------------------

#ifdef WITH_DAEMON

int syslog_log_priorities[] = {
    [LOG_FATAL] = LOG_CRIT,
    [LOG_ERR]   = LOG_ERR,
    [LOG_MSG]   = LOG_WARNING,
    [LOG_INF]   = LOG_INFO,
    [LOG_DEB]   = LOG_DEBUG,
    [LOG_TRASH] = LOG_DEBUG,
};

void
syslog_log_fnc(const char *domain, int level,
	const char *format, va_list ap)
{
    char *str;

    level = syslog_log_priorities[level & VCA_LOGL_MASK];
    if(vasprintf(&str, format, ap)<0)
        return;
    if(domain == NULL)
        syslog(LOG_MAKEPRI(LOG_LOCAL5, level), "%s", str);
    else
        syslog(LOG_MAKEPRI(LOG_LOCAL5, level), "%s:%s", domain, str);
    free(str);
}


static int daemon_init(void) {
    pid_t pid;

    if ((pid = fork()) < 0) {
        return -1;
    } else {
        if (pid != 0) {
            exit(0);	/* parent vanishes */
        }
    }
    /* child process */
    setsid();
    umask(022);

    if(getenv("VCA_LOG_FILENAME")==NULL) {
        openlog ("canblaster", 0, LOG_LOCAL3);
        vca_log_redir(syslog_log_fnc, 0);
    }

    close(0);
    close(1);
    close(2);
    return 0;
}

#endif /*WITH_DAEMON*/

//----------------------------------------------------

int main(int argc, char *argv[])
{
    int terminate_enable_fl = 0;
    int terminate_fl=0;
    int daemon_fl=0;
    int help_fl=0;

    vcaFdPoll_t fdpoll;
    vcaFdPoll_init(&fdpoll);
    
    /* initialize hierarchical timers */
    ul_mstime_init();
    ul_htimer_init_queue(&ul_root_htimer);
    
    /* parse options */
	printf("CANBLASTER - CANopen simple master\n");
    {
        #ifdef DEBUG
        vca_log_cutoff_level = LOG_DEB;
        #else
        vca_log_cutoff_level = LOG_MSG;
        #endif
        /* parse cmd line arguments */
        while(1) { 
            static struct option long_options[] = {
                {"port", required_argument, 0, 'p'},
                {"candev", required_argument, 0, 'd'},
                {"verbose", required_argument, 0, 'v'},
		{"terminate-from-tty", no_argument, 0, 't'},
              #ifdef WITH_DAEMON
                {"daemon", no_argument, 0, 'D'},
              #endif /*WITH_DAEMON*/
                {"log_level", required_argument, 0, 'g'},
                {"help", no_argument, 0, 'h'},
                {0, 0, 0, 0}
            };
            /* getopt_long stores the option index here. */
            int option_index = 0;
            int c;
            c = getopt_long(argc, argv, "d:p:g:hv", long_options, &option_index);
            /* Detect the end of the options. */
            if (c == -1) break;
            switch (c) {
                case 0:
                    /*
                    printf ("option %s", long_options[option_index].name);
                    if (optarg) printf (" with arg %s", optarg);
                    printf ("\n"); */
                    break;
                case 'p':
                    server_port = atoi(optarg);
                    clog(LOG_DEB, "option -p with value `%s'\n", optarg);
                    break;
                case 'g':
                    vca_log_cutoff_level = atoi(optarg);
                    clog(LOG_DEB, "option -g with value `%s'\n", optarg);
                    break;
                case 'd':
                    can_device_name = optarg;
                    break;
                case 'v':
                    vca_log_cutoff_level = LOG_INF;
                    break;
                case 't':
                    terminate_enable_fl = 1;
                    break;
                case 'D':
                    daemon_fl = 1;
                    break;
                case 'h':
                    help_fl = 1;
                    clog(LOG_DEB, "option -h\n");
                    break;
                case '?':
                    /* getopt_long already printed an error message. */
                    break;
                default:
                    fatal_error(LOG_FATAL, "cmd line parsing error\n");
            }
        }
        if(help_fl) {
            printf(help_msg, argv[0]);
            exit(EXIT_SUCCESS);
        } 
        /* Print any remaining command line arguments (not options). */
        if (optind < argc) {
            clog(LOG_INF, "non-option ARGV-elements: ");
            while (optind < argc) clog(LOG_INF, "%s ", argv[optind++]);
            clog(LOG_INF, "\n");
        }
        clog(LOG_INF, "Debug level: %i\n", vca_log_cutoff_level);
    }
    
#ifdef WITH_DAEMON
    if(daemon_fl) {
        daemon_init();
    }
#endif /*WITH_DAEMON*/

    /* init CAN net and register fd into vcaFdPoll */
    {
        vcaNet_init(&can_net);
        vcaNet_connect_device(&can_net, NULL, can_device_name);
        vcaFdPoll_insert(&fdpoll, &(can_net.canfdnode->fdnode));
    }
    
    /* open server socket and register fd into vcaFdPoll */
    {
        struct sockaddr_in server_address;
        int server_len;
        int yes = 1;
        server_address.sin_family = AF_INET;
        server_address.sin_addr.s_addr = htonl(INADDR_ANY);
        server_address.sin_port = htons(server_port);
        server_len = sizeof(server_address);
    
        FdNodeSrvSock_init(&server_node);
        server_node.fdnode.fd = socket(AF_INET, SOCK_STREAM, 0);
        clog(LOG_TRASH, "server fd: %i\n", server_node.fdnode.fd);
        if(server_node.fdnode.fd < 0) {
            fatal_error("ERROR open server socket\n");
        }
        if(fcntl(server_node.fdnode.fd, F_SETFL, O_NONBLOCK) < 0) {
            fatal_error("ERROR fcntl\n");
        }
        if(setsockopt(server_node.fdnode.fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes)) == -1) {
            fatal_error("ERROR setsockopt\n");
        }
        if(bind(server_node.fdnode.fd, (struct sockaddr *)&server_address, server_len) < 0) {
            fatal_error("ERROR binding server socket\n");
        }
        if(listen(server_node.fdnode.fd, 10)) {
            fatal_error("ERROR listen in server socket\n");
        }
        clog(LOG_INF, "Listenning on port %i\n", server_port);
        
        server_node.cannet = &can_net;
        vcaFdPoll_insert(&fdpoll, &(server_node.fdnode));
    }


    /* Support for termination by standard input*/
    if(terminate_enable_fl) {
        cmdttyFdNode_init(&cmdttyFdNode,0,&terminate_fl);
        vcaFdPoll_insert(&fdpoll, &(cmdttyFdNode.fdnode));
    }

    do {
        ul_msdiff_t ms_sleep;
        ul_mstime_t ms_last, ms_next;
        ul_mstime_now(&ms_last);
        ul_htimer_run_expired(&ul_root_htimer, &ms_last);
        if(!ul_htimer_next_expire(&ul_root_htimer, &ms_next)) {
            ms_sleep = -1;
        }
        else {
            ul_mstime_now(&ms_last);
            ms_sleep = ms_next - ms_last;
            if(ms_sleep < 0) ms_sleep=0;
        }
        //log(LOG_TRASH, "pull in\n");
        vcaFdPoll_poll(&fdpoll, ms_sleep);
        //log(LOG_TRASH, "pull out\n");
        
    } while(!terminate_fl);
    
    /* optional data structures cleanup */
    clean_up();

    vcaFdPoll_destroy(&fdpoll);

    return 0;
}
